
import json
import collections
import typing
from matplotlib import pyplot as plt
import math
import os
import pandas as pd

from twitch_chat_analyzer import kraken
from twitch_chat_analyzer import downloader
from twitch_chat_analyzer import models as tca_models


class ChatAnalyzer:
  def __init__(self, video_json, comments_json):
    self.video = video_json
    self.comments = [tca_models.Comment(comment_json) for comment_json in comments_json]

  def DrawChatPerMinute(self, minute: int = 5):
    chat_counts = self.GetChatPerMinute(minute)
    x = [str(index * minute) for index in range(len(chat_counts))]

    plt.bar(x, chat_counts, color='pink')
    plt.show()

  def GetChatPerMinute(self, minute: int = 5) -> typing.List[int]:
    total_seconds = minute * 60
    total_slots = int(math.ceil(self.video['length'] / total_seconds))
    counts = [0] * total_slots
    for comment in self.comments:
      index = int(comment.offset / total_seconds)
      counts[index] += 1

    return counts

  def DrawTopEmotes(self, top: int = 10):
    emote_counts = self.GetEmoteCounts()[:top]
    names, counts = zip(*emote_counts)

    plt.xticks(rotation=45, ha='right')
    plt.bar(names, counts, color='#00917C')
    plt.show()

  def GetEmoteCounts(self) -> typing.List[typing.Tuple[str, int]]:
    emote_dict: typing.Dict[str, int] = collections.defaultdict(int)
    for comment in self.comments:
      emotes = comment.emotes
      for emote in emotes:
        emote_dict[emote] += 1
    
    return self._SortedByCount(emote_dict)
  
  def DrawTopUniqueEmotes(self, top: int = 10):
    unique_emote_counts = self.GetUniqueEmoteCounts()[:top]
    names, counts = zip(*unique_emote_counts)

    plt.xticks(rotation=45, ha='right')
    plt.bar(names, counts, color='#61B15A')
    plt.show()

  def GetUniqueEmoteCounts(self) -> typing.List[typing.Tuple[str, int]]:
    unique_emote_dict: typing.Dict[str, int] = collections.defaultdict(int)
    for comment in self.comments:
      unique_emotes = comment.unique_emotes
      for emote in unique_emotes:
        unique_emote_dict[emote] += 1

    return self._SortedByCount(unique_emote_dict)

  def DrawTopChatters(self, top: int = 10):
    chat_counts = self.GetTopChatters()[:top]
    names, counts = zip(*chat_counts)

    plt.xticks(rotation=45, ha='right')
    plt.bar(names, counts, color='#79A3B1')
    plt.show()

  def GetTopChatters(self) -> typing.List[typing.Tuple[str, int]]:
    chat_count: typing.Dict[str, int] = collections.defaultdict(int)
    for comment in self.comments:
      chat_count[comment.display_name] += 1

    return self._SortedByCount(chat_count)

  def ToDataFrame(self) -> pd.DataFrame:
    return pd.DataFrame([comment.ToDict() for comment in self.comments])  

  def _SortedByCount(self, counts: typing.Dict[str, int]) -> typing.List[typing.Tuple[str, int]]:
    sorted_counts = sorted(counts.items(), key=lambda item: (-item[1], item[0]))
    return sorted_counts


def FromFile(filepath) -> ChatAnalyzer:
  with open(filepath, 'r', encoding='utf8') as f:
    try:
      data_json = json.load(f)
      return ChatAnalyzer(data_json['video'], data_json['comments'])
    except Exception as e:
      print('Cannot load the file', filepath, '. Reason:', e)
      raise e


def FromVideoId(video_id) -> ChatAnalyzer:
  # First check if the chat log was already downloaded
  client = kraken.TwitchClient()
  username = client.GetUsernameFromVideo(video_id)
  path = 'chatlogs/{username}/{video_id}.json'.format(username=username, video_id=video_id)
  if os.path.exists(path):
    return FromFile(path)
  
  # If not already downloaded, download the chat
  data_json = downloader.downloadChat(video_id)
  return ChatAnalyzer(data_json['video'], data_json['comments'])




import typing
import requests

BASE_URL = 'https://api.twitch.tv/v5/%s'
USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0'


CLIENT_ID = 's2frr164040bp9l6mgurxaz32i6rm1'

class TwitchClient:

  def __init__(self, client_id: str = CLIENT_ID):
    self.client_id = client_id
    self.session = requests.Session()
    self.headers = {
      'Client-ID': client_id,
      'Accept': 'application/vnd.twitchtv.v5+json'
    }

  def GetUsernameFromVideo(self, video_id) -> str:
    video_json = self.GetVideo(video_id)
    if not video_json:
      return None

    return video_json['channel']['name']

  def GetVideo(self, video_id) -> dict[str, typing.Any]:
    url = BASE_URL % 'videos/' + str(video_id)
    return self._GetUrl(url)

  def GetComments(self, video_id):
    cursor = ''
    while cursor is not None:
      url = BASE_URL % 'videos/' + str(video_id) + '/comments?cursor=' + cursor
      comments_batch = self._GetUrl(url)
      cursor = comments_batch.get('_next')
      yield comments_batch['comments']

  def _GetUrl(self, url):
    response = self.session.get(url, headers=self.headers)
    if response.status_code >= 400:
      print('Request to', url, 'failed with status code', response.status_code)
      return None

    try:
      return response.json()
    except Exception as e:
      print('JSON parsing error', e)
      return None



import time
import sys
import os
import json

from twitch_chat_analyzer import kraken

CLIENT_ID = "s2frr164040bp9l6mgurxaz32i6rm1"  # Test client ID 


def downloadChat(video_id, save_to_file=True):
  client = kraken.TwitchClient(CLIENT_ID)
  video_json = client.GetVideo(video_id)
  if not video_json:
    return
  # print(json.dumps(video_json, indent=2))

  print('Downloading chat of', video_json['title'], 'by', video_json['channel']['display_name'])

  total_seconds = float(video_json['length'])

  start_time = time.time()
  comments_generator = client.GetComments(video_id)
  comments = []
  for comments_batch in comments_generator:
    comments.extend(comments_batch)
    comment_count = len(comments)
    if comment_count:
      draw_progress(comments[-1]['content_offset_seconds'], total_seconds, comment_count)

  draw_progress(100, 100, len(comments))
  end_time = time.time()
  sys.stdout.write('\nDownload complete in %d seconds' % (end_time - start_time))

  data_json = {"video": video_json, "comments": comments}
  if save_to_file:
    save_chat(video_id, data_json)

  return data_json


def draw_progress(offset, total, comment_count):
  progress = min(100, offset / total * 100)
  sys.stdout.write('Total %d chats downloaded. %.2f%% complete\r' % (comment_count, progress))
  sys.stdout.flush()

def save_chat(video_id, data_json):
  video_json = data_json['video']
  streamer_username = video_json['channel']['name']
  path_dir = os.path.join('.', 'chatlogs', streamer_username)
  if not os.path.exists(path_dir):
    os.makedirs(path_dir)
  filepath = os.path.join(path_dir, str(video_id) + '.json')
  content = json.dumps(data_json, indent=2, ensure_ascii=False)
  with open(filepath, 'w', encoding='utf8') as f:
    f.write(content)
  print('\nChat saved to ', filepath, '\n')


if __name__ == '__main__':
  data = downloadChat('past_broadcast_number')




'''Model classes'''

import typing


# Model class for chat comment
class Comment:
  def __init__(self, comment_json):
    self._comment = comment_json
    self._commenter = comment_json['commenter']
    self._message = comment_json['message']
    self._fragments = self._message['fragments']

  @property
  def offset(self) -> float:
    return self._comment['content_offset_seconds']

  @property
  def username(self) -> str:
    # Twitch login username
    return self._commenter['name']
  
  @property
  def display_name(self) -> str:
    # Twitch display name in chat, which may not be in English
    return self._commenter['display_name']

  @property
  def name(self) -> str:
    # Name as is it shown in chat.
    # For display names with English alphabets, this is the same as display name.
    # For non-English display names, this is '{display_name}({username})'
    username = self.username
    display_name = self.display_name
    if username.lower() == display_name.lower():
      return display_name

    return '%s(%s)' % (display_name, username)

  @property
  def body(self) -> str:
    # Raw chat content
    return self._message['body']

  @property
  def text_body(self) -> str:
    # Chat text content excluding emotes
    texts = []
    for fragment in self._fragments:
      if 'emoticon' not in fragment:
        texts.append(fragment['text'].strip())

    return ' '.join(texts)
  
  @property
  def emotes(self) -> typing.List[str]:
    emotes = []
    for fragment in self._fragments:
      if 'emoticon' in fragment:
        emotes.append(fragment['text'])

    return emotes

  @property
  def unique_emotes(self) -> typing.Set[str]:
    return set(self.emotes)

  @property
  def is_subscriber(self) -> bool:
    badges = self._message.get('user_badges', [])
    for badge in badges:
      if badge['_id'] == 'subscriber':
        return True
    
    return False

  @property
  def bits(self) -> int:
    return self._message.get('bits_spent', 0)
  
  @property
  def is_sub_notice(self) -> bool:
    # If this is a subscription notice message
    msg_id = self._message.get('user_notice_params', {}).get('msg-id')
    return msg_id in ('sub', 'resub')

  def ToDict(self) -> typing.Dict[str, typing.Union[str, int, float]]:
    return {
      'offset': self.offset,
      'username': self.username,
      'display_name': self.display_name,
      'name': self.name,
      'body': self.body,
      'text_body': self.text_body,
      'is_subscriber': self.is_subscriber,
      'bits': self.bits,
      'is_sub_notice': self.is_sub_notice,
    }